"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("mz/fs");
const path = require("path");
const cancellation_1 = require("./cancellation");
const glob = require("glob");
const semaphore_async_await_1 = require("semaphore-async-await");
const url = require("url");
const util_1 = require("./util");
class RemoteFileSystem {
    constructor(client) {
        this.client = client;
    }
    /**
     * The files request is sent from the server to the client to request a list of all files in the workspace or inside the directory of the base parameter, if given.
     * A language server can use the result to index files by filtering and doing a content request for each text document of interest.
     */
    getWorkspaceFiles(base, token = cancellation_1.CancellationToken.None) {
        return __awaiter(this, void 0, void 0, function* () {
            const textDocuments = yield this.client.getWorkspaceFiles({ base }, token);
            return textDocuments.map(textDocument => textDocument.uri);
        });
    }
    /**
     * The content request is sent from the server to the client to request the current content of any text document. This allows language servers to operate without accessing the file system directly.
     */
    getTextDocumentContent(uri, token = cancellation_1.CancellationToken.None) {
        return __awaiter(this, void 0, void 0, function* () {
            const textDocument = yield this.client.getTextDocumentContent({ textDocument: { uri } }, token);
            return textDocument.text;
        });
    }
}
exports.RemoteFileSystem = RemoteFileSystem;
class LocalFileSystem {
    /**
     * @param rootPath The root directory path that relative URIs should be resolved to
     */
    constructor(rootPath) {
        this.rootPath = rootPath;
    }
    /**
     * Converts the URI to an absolute path
     */
    resolveUriToPath(uri) {
        return path.resolve(this.rootPath, util_1.uri2path(uri));
    }
    getWorkspaceFiles(base) {
        return __awaiter(this, void 0, void 0, function* () {
            const pattern = base ? path.posix.join(this.resolveUriToPath(base), '**/*.*') : this.resolveUriToPath('file:///**/*.*');
            const files = yield new Promise((resolve, reject) => {
                glob(pattern, { nodir: true }, (err, matches) => err ? reject(err) : resolve(matches));
            });
            return files.map(file => util_1.path2uri('', file));
        });
    }
    getTextDocumentContent(uri) {
        return __awaiter(this, void 0, void 0, function* () {
            return fs.readFile(this.resolveUriToPath(uri), 'utf8');
        });
    }
}
exports.LocalFileSystem = LocalFileSystem;
/**
 * Memoization cache that saves URIs and searches parent directories too
 */
class ParentUriMemoizationCache extends Map {
    /**
     * Returns the value if the given URI or a parent directory of the URI is in the cache
     */
    get(uri) {
        let hit = super.get(uri);
        if (hit) {
            return hit;
        }
        // Find out if parent folder is being fetched already
        hit = super.get(undefined);
        if (hit) {
            return hit;
        }
        if (uri) {
            for (let parts = url.parse(uri); parts.pathname && parts.pathname !== '/'; parts.pathname = path.dirname(parts.pathname)) {
                hit = super.get(url.format(parts));
                if (hit) {
                    return hit;
                }
            }
        }
        return undefined;
    }
    /**
     * Returns true if the given URI or a parent directory of the URI is in the cache
     */
    has(key) {
        return this.get(key) !== undefined;
    }
}
/**
 * Synchronizes a remote file system to an in-memory file system
 *
 * TODO: Implement Disposable with Disposer
 */
class FileSystemUpdater {
    constructor(remoteFs, inMemoryFs) {
        this.remoteFs = remoteFs;
        this.inMemoryFs = inMemoryFs;
        /**
         * Limits concurrent fetches to not fetch thousands of files in parallel
         */
        this.concurrencyLimit = new semaphore_async_await_1.default(100);
        /**
         * Returns a promise that is resolved when the given URI has been fetched (at least once) to the in-memory file system.
         * This function cannot be cancelled because multiple callers get the result of the same operation.
         *
         * @param uri URI of the file to ensure
         */
        this.ensure = cancellation_1.cancellableMemoize(this.fetch);
        /**
         * Returns a promise that is resolved as soon as the file/directory structure for the given directory has been synced
         * from the remote file system to the in-memory file system (at least once)
         */
        this.ensureStructure = cancellation_1.cancellableMemoize(this.fetchStructure, new ParentUriMemoizationCache());
    }
    /**
     * Fetches the file content for the given URI and adds the content to the in-memory file system
     *
     * @param uri URI of the file to fetch
     */
    fetch(uri, token = cancellation_1.CancellationToken.None) {
        // Limit concurrent fetches
        return this.concurrencyLimit.execute(() => __awaiter(this, void 0, void 0, function* () {
            cancellation_1.throwIfRequested(token);
            const content = yield this.remoteFs.getTextDocumentContent(uri, token);
            this.inMemoryFs.add(uri, content);
        }));
    }
    /**
     * Fetches the file/directory structure for the given directory from the remote file system and saves it in the in-memory file system
     *
     * @param base The base directory which structure will be synced. Defaults to the workspace root
     */
    fetchStructure(base, token = cancellation_1.CancellationToken.None) {
        return __awaiter(this, void 0, void 0, function* () {
            const uris = yield this.remoteFs.getWorkspaceFiles(base, token);
            for (const uri of uris) {
                this.inMemoryFs.add(uri);
            }
        });
    }
}
exports.FileSystemUpdater = FileSystemUpdater;
//# sourceMappingURL=fs.js.map